use axum::{
    http::header::SET_COOKIE,
    response::{AppendHeaders, IntoResponse, Redirect},
    routing::post,
    Extension, Form, Router,
};

use lettre::message::{header::ContentType, Mailbox, SinglePart};
use lettre::transport::smtp::AsyncSmtpTransport;
use lettre::{AsyncTransport, Message, Tokio1Executor};
use serde::{Deserialize, Deserializer};
use std::error::Error;
use std::net::SocketAddr;
use std::sync::Arc;

fn deserialize_empty_string_as_none<'de, D, T>(de: D) -> Result<Option<T>, D::Error>
where
    D: Deserializer<'de>,
    T: std::str::FromStr + std::fmt::Debug,
    T::Err: std::fmt::Display + std::fmt::Debug,
{
    let opt: Option<String> = Option::deserialize(de)?;
    match opt.as_deref() {
        None | Some("") => Ok(None),
        Some(s) => std::str::FromStr::from_str(s)
            .map_err(serde::de::Error::custom)
            .map(Some),
    }
}

#[derive(Deserialize)]
#[serde(rename_all = "kebab-case")]
struct FormData {
    #[serde(default, deserialize_with = "deserialize_empty_string_as_none")]
    name: Option<String>,
    email: String,
    #[serde(default, deserialize_with = "deserialize_empty_string_as_none")]
    company_name: Option<String>,
    service: String,
    help: String,
    #[serde(default, deserialize_with = "deserialize_empty_string_as_none")]
    hear_about: Option<String>,
    #[serde(default, deserialize_with = "deserialize_empty_string_as_none")]
    other_source: Option<String>,
}

struct State {
    username: String,
    password: String,
    target_address: String,
    cc: Option<String>,
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let username = std::env::var("WEBMAIL_USERNAME")?;
    let password = std::env::var("WEBMAIL_PASSWORD")?;
    let target_address = std::env::var("WEBMAIL_TARGET")?;
    let cc = std::env::var("WEBMAIL_CC").ok();

    let state = State {
        username,
        password,
        target_address,
        cc,
    };

    let app = Router::new()
        .route("/submit-email", post(handle_form))
        .layer(Extension(Arc::new(state)));

    let addr = SocketAddr::from(([0, 0, 0, 0], 4200));
    println!("Listening on {}", addr);

    let listener = tokio::net::TcpListener::bind("0.0.0.0:4200").await?;
    axum::serve(listener, app).await?;

    Ok(())
}

async fn handle_form(
    state: Extension<Arc<State>>,
    Form(form_data): Form<FormData>,
) -> impl IntoResponse {
    match send_email(form_data, state.as_ref()).await {
        Ok(_) => (
            AppendHeaders([(SET_COOKIE, "is=ok")]),
            Redirect::to("/contact.html"),
        ),
        Err(e) => {
            eprintln!("Encountered error sending mail: {e:?}");
            (
                AppendHeaders([(SET_COOKIE, "is=err")]),
                Redirect::to("/contact.html"),
            )
        }
    }
}

async fn send_email(form_data: FormData, state: &State) -> Result<(), Box<dyn Error>> {
    let username = &state.username;
    let password = &state.password;

    let email = form_data.email.parse::<Mailbox>()?;

    let content = [
        (
            "Name?",
            form_data.name.unwrap_or("Not provided".to_string()),
        ),
        ("Email", form_data.email),
        (
            "Company Name?",
            form_data.company_name.unwrap_or("Not provided".to_string()),
        ),
        ("Service", form_data.service),
        ("How can we help?", form_data.help),
        (
            "How did you hear about us?",
            form_data.hear_about.unwrap_or("Not provided".to_string()),
        ),
        (
            "Other",
            form_data
                .other_source
                .unwrap_or("Not provided/relevant".to_string()),
        ),
    ];
    let body = content.map(|(a, b)| format!("{a}: {b}")).join("\n");

    let mut email_builder = Message::builder()
        .from(username.parse()?)
        .reply_to(email)
        .to(state.target_address.parse()?)
        .subject("New Website Form Inquiry");

    if let Some(cc) = state.cc.as_ref() {
        email_builder = email_builder.cc(cc.parse()?);
    }

    let email = email_builder.singlepart(
        SinglePart::builder()
            .header(ContentType::TEXT_PLAIN)
            .body(body),
    )?;

    let credentials = lettre::transport::smtp::authentication::Credentials::new(
        username.clone(),
        password.clone(),
    );

    let mailer = AsyncSmtpTransport::<Tokio1Executor>::relay("smtp.migadu.com")?
        .credentials(credentials)
        .build();

    let response = mailer.send(email).await?;
    dbg!(&response);

    Ok(())
}
